//A Stitchway behaves like a normal Oogway in the sense that it
//produces lines on the screen and in a pdf file, when being recorded, but it also
//produces stitches in a .dst file. Restrictions apply such as no splines and no colors.
//The Stitchway also has a beginRecord (now called beginRecordTajima) which does both dst and pdf.
//Syntax description of file taken from KDE Community Wiki, "Tajima Ternary DST", "Header" and "Stitch".
//Coding of stitches adapted from：
//github.com/Embroidermodder/Embroidermodder/blob/master/libembroidery/format-dst.c
//Attention: there is a scaling factor between pixels (Oogway's unit) and millimeters (related to the 0.1mm of Tajima files.
//Attention: the Oogway screen image and the Tajima files appear mirrored along the x axis. 
//Need constant float PIXELSPERMILLIMETER = 3.6;
//(C) Loe Feijs and TU/e 2014 

//Addition 2016: new command, nextColor(), which moves through a 6-entries color table
//(darkblue, lightblue, green, yellow, bordeaux, brown). In .DST it generates Tajima 00 00 C3
//"Color Change" command. And in .PDF it generates the colors from the abovementioned table,
//which simulates the behaviour of the .DST file in Wilcom TrueSizer (at least for six colors).
//The classical oogway command setPenColor() also works for .PDF, but not for .DST.

float PIXELSPERMILLIMETER = DPMM;

class Stitchway {
      Oogway o;
      int MINX, MAXX;//units of .1mm
      int MINY, MAXY;//idem
      int xneedle;//embroidery needle, like xcor, but in .1mm units 
      int yneedle;//idem
      int stitchcount; 
      Stitch[] stitches;
      int bytecount;
      byte[] bytes; 
      PGraphicsPDF pdf;
      String filename; 
      int MAXNRSTITCHES = 10 * 1000 * 1000;//eg 2 * 1000 * 1000
      int currentColor = 0; //points into wilcom table
      color[] wilcom = new color[6];

      Stitchway(PApplet p, int MINX, int MAXX, int MINY, int MAXY){
             o = new Oogway(p);
             o.setPosition(0,0);
             this.MINX = MINX; this.MAXX = MAXX;
             bytecount = 0;
             bytes = new byte[512 + 3*MAXNRSTITCHES + 1]; 
             stitchcount = 0;//HERSTEL NAAR 0
             stitches = new Stitch[MAXNRSTITCHES];
             xneedle = floor(10 * o.xcor());
             yneedle = floor(10 * o.ycor());
             wilcom[0] = color(86,54,122);   //darkblue
             wilcom[1] = color(102,199,223); //lightblue
             wilcom[2] = color(0,176,0);     //green
             wilcom[3] = color(240,234,4);   //yellow
             wilcom[4] = color(194,84,78);   //bordeaux;
             wilcom[5] = color(190,132,73);  //brown    
             o.setPenColor(wilcom[currentColor]);             
      }
      //only a few commands implemented at Tajima:
      //(PS the screen and pdf turtle work anyway)
      void setPosition(float x, float y){
           o.setPosition(x,y);
           int newxneedle = floor(10 * o.xcor()/PIXELSPERMILLIMETER);
           int newyneedle = floor(10 * o.ycor()/PIXELSPERMILLIMETER);
           Stitch s = new Stitch(newxneedle - xneedle, newyneedle - yneedle,true);
           s.mkMultiStitch(this);
           xneedle = newxneedle;
           yneedle = newyneedle;
      }
      void forward(float distance){
           o.forward(distance);
           int newxneedle = floor(10 * o.xcor()/PIXELSPERMILLIMETER);
           int newyneedle = floor(10 * o.ycor()/PIXELSPERMILLIMETER);
           Stitch s = new Stitch(newxneedle - xneedle, newyneedle - yneedle, !o.isDown());
           s.mkMultiStitch(this);
           xneedle = newxneedle;
           yneedle = newyneedle;
      }
      void nextColor(){
           //embroidering first
           Stitch s = new Stitch(0,0,false);
           s.jumpstitch = false;
           s.colourstop = true;
           s.endpattern = false;
           s.dx = 0;
           s.dy = 0;
           this.stitches[this.stitchcount++] = s;
           //pdf color secondly
           if (currentColor < 5)
               o.setPenColor(wilcom[++currentColor]);
      }
      //and a few "inherited" from Oogway
      void penUp(){
           o.penUp(); 
      }
      void penDown(){
           o.penDown(); 
      }
      float xcor(){
            return o.xcor();
      }
      float ycor(){
            return o.ycor();
      }
      void left(float angle){
           o.left(angle);
      }
      void right(float angle){
           o.right(angle);
      }
      void setPenSize(float size){
           o.setPenSize(size);
      }
      float penSize(){
           return o.penSize();
      }
      void setHeading(float to_angle){
           o.setHeading(to_angle);
      }
      void setPenColor(int red, int green, int blue){
           o.setPenColor(red, green, blue);
      }
      color penColor(){
            return o.penColor();
      }
      boolean isDown(){
              return o.isDown();
      }
      //and a few derived commands:
      void setPos(float x, float y){
           setPosition(x,y);
      }
      void fd(float distance){
           forward(distance); 
      }
      void lt(float angle){
           left(angle);
      }
      void rt(float angle){
           right(angle);
      }
      private void mkField(int offset, int size, String label, String text){
              if (label.length() != 3) 
                  println("ERROR HEADER LABEL");
              for (int i = offset; i < offset + size; i++)
                      bytes[i] = ' ';
              for (int i = 0; i < 3 ; i++)
                   bytes[offset + i] = byte(label.charAt(i));
              for (int i = 0; i < min(text.length(),size - 4); i++)
                   bytes[offset + 3 + i] = byte(text.charAt(i));    
              byte CR = 13; //^M, carriage return (?)   
              bytes[offset + size - 1] = CR;
      }
      private void mkField(int offset, int size, String label, int number){
              if (label.length() != 3) 
                  println("ERROR HEADER LABEL"); 
              if (stitchcount > 2147483647)
                  println("ERROR NUMBER EXCEEDS MAXINT"); 
              for (int i = offset; i < offset + size; i++)
                      bytes[i] = ' ';
              for (int i = 0; i < 3; i++)
                   bytes[offset + i] = byte(label.charAt(i));  
              int nrdigits = int(1 + (float)Math.log10(max(1,number)));
              if (3 + nrdigits + 1 > size) 
                  println("ERROR NUMBER SIZE EXCEEDS FIELD SIZE"); 
              String text = nf(number,nrdigits);
              for (int i = 0; i < nrdigits; i++)
                   bytes[offset + size - 1 - nrdigits + i] = byte(text.charAt(i));    
              byte CR = 13; //^M, carriage return (?) 
              bytes[offset + size - 1] = CR; 
      }
      private void mkFieldPlus(int offset, int size, String label, int number){
              //similar to mkField but adds a "+" after the ":"
              if (label.length() != 3) 
                  println("ERROR HEADER LABEL"); 
              if (stitchcount > 2147483647)
                  println("ERROR NUMBER EXCEEDS MAXINT"); 
              for (int i = offset; i < offset + size; i++)
                      bytes[i] = ' ';
              for (int i = 0; i < 3; i++)
                   bytes[offset + i] = byte(label.charAt(i)); 
              bytes[offset + 3] = '+'; 
              int nrdigits = int(1 + (float)Math.log10(max(1,number)));
              if (4 + nrdigits + 1 > size) 
                  println("ERROR NUMBER SIZE EXCEEDS FIELD SIZE"); 
              String text = nf(number,nrdigits);
              for (int i = 0; i < nrdigits; i++)
                   bytes[offset + size - 1 - nrdigits + i] = byte(text.charAt(i));    
              byte CR = 13; //^M, carriage return (?) 
              bytes[offset + size - 1] = CR; 
      }
      void mkHeader(String title){
           for (int i = 0; i<512; i++)
                bytes[i] = ' ';
           //offset 0, size 20, ASCII = "LA:" (label) Name padded with spaces 0x20, sometimes terminated with nul 0x00
           //Offset 20, Size 11, ASCII = "ST:" (stitch count) decimal number padded with zero or space
           //Offset 31, Size 7, ASCII = "CO:" (colour count) decimal number
           //Offset 38, Size 9, ASCII = "+X:" (maximum X extent) decimal number in tenths of a millimetre
           //Offset 47, Size 9, ASCII = "-X:" (minimum X extent) decimal number in tenths of a millimetre
           //Offset 56, Size 9, ASCII = "+Y:" (maximum Y extent) decimal number in tenths of a millimetre
           //Offset 65, Size 9, ASCII = "-Y:" (minimum Y extent) decimal number in tenths of a millimetre
           //Offset 74, Size 10, ASCII = "AX:" (needle end point X) "+" or "-" decimal number in tenths of a millimetre
           //Offset 84, Size 10, ASCII = "AY:" (needle end point Y) "+" or "-" decimal number in tenths of a millimetre
           //Offset 94, Size 10, ASCII = "MX:" (previous file needle end point X) "+" or "-" decimal number in tenths of a millimetre
           //Offset 104, Size 10, ASCII = "MY:" (previous file needle end point Y) "+" or "-" decimal number in tenths of a millimetre
           //Offset 114, Size 10, ASCII = "PD:" (previous file) "******" if single file
           //Offset 124, Size 4, ASCII = 0x1a 0x20 0x20 0x20 or 0x1a 0x00 0x00 0x00 end of header data
           //Offset 128, Size 384, ASCII = spaces 0x20 or sometimes contains colour information  
           byte CR = 13; //^M, carriage return (?) 
           mkField( 0,20,"LA:",title);
           mkField(20,11,"ST:",stitchcount);
           mkField(31,7,"CO:",0);
           mkField(38,9,"+X:",int(10 * PIXELSPERMILLIMETER * 210));//ASSUMPTION: A4 LANDSCAPE
           mkField(47,9,"-X:",0);
           mkField(56,9,"+Y:",int(10 * PIXELSPERMILLIMETER * 297));//ASSUMPTION: A4 LANDSCAPE
           mkField( 65, 9,"-Y:",0);//not implemented
           mkFieldPlus( 74,10,"AX:",0);//not implemented
           mkFieldPlus( 84,10,"AY:",0);//not implemented
           mkFieldPlus( 94,10,"MX:",0);//not implemented
           mkFieldPlus(104,10,"MY:",0);//not implemented
           mkField(114,10,"PD:","******");
           byte SUB = 26;//^Z aka byte(\u001A)
           bytes[124] = SUB;
           bytecount = 512;      
      }
      void beginRecordTajima(String fn){
           //fn defines both pdf and dst file
           filename = fn;
           pdf = (PGraphicsPDF) beginRecord(PDF, filename + ".pdf");
           //we do the dst later, at endRecord
      }
      void endRecordTajima(){
           endRecord();
           //finalise header bytes:
           mkHeader(filename);
           //finalise stitches bytes:
           for (int i = 0; i < stitchcount; i++)
                stitches[i].insert(this);
           //add one more "endofpattern" stitch:
           Stitch s = new Stitch(0,0,false);
           s.endpattern = true;
           s.insert(this);
           //repack everything for saving:
           byte[] finalbytes = new byte[512 + 3*stitchcount + 1]; 
           for (int i = 0; i < 512 + 3*stitchcount; i++)
                finalbytes[i] = bytes[i];
           //add a final ^Z and save:
           byte SUB = 26;
           finalbytes[512 + 3*stitchcount] = SUB;
           saveBytes(filename + ".dst", finalbytes);
      }
}//end class Tajima
